import functools
import threading
from copy import deepcopy


def closure(f):
    def closure_aux(item, *args):
        prec = None
        while item != prec:
            prec = deepcopy(item)
            item = f(item, *args)
        return item

    return closure_aux


def preconditions(*members, msg_to_print=None, valid=lambda *params: all(p is not None for p in params)):
    def wrapper_method(method):
        def validation(ref, *args, **kwargs):
            if not valid(*(getattr(ref, member) for member in members)):
                msg = msg_to_print if msg_to_print else f"Violated prerequisites of method {method.__name__}"
                raise Exception(msg)
            return method(ref, *args, **kwargs)

        return validation

    return wrapper_method


def synchronized(lock: threading.RLock):
    def _decorator(wrapped):
        @functools.wraps(wrapped)
        def _wrapper(*args, **kwargs):
            with lock:
                return wrapped(*args, **kwargs)

        return _wrapper

    return _decorator


class MetaClassSingleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in MetaClassSingleton._instances:
            MetaClassSingleton._instances[cls] = super().__call__(*args, **kwargs)
        return MetaClassSingleton._instances[cls]


class DecoratorSingleton:
    def __init__(self, decorated_type):
        self._decorated_type = decorated_type
        self._instance = None

    @property
    @preconditions("_instance")
    def instance(self):
        return self._instance

    def __call__(self, *args, **kwargs):
        if self._instance is not None: raise TypeError(
            'You already had instantiate this object, please call instance() method to retrieve')
        self._instance = self._decorated_type(*args, **kwargs)
        return self._instance

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated_type)

    def destroy(self):
        del self._instance
        self._instance = None



import abc
import enum

from pycarlanet import CarlanetActor

"""
Listener for OMNeT world, note, every parameters gave as in input is a String, so eventually you need to use cast
"""


class SimulatorStatus(enum.Enum):
    RUNNING = 0
    FINISHED_OK = 1
    FINISHED_ERROR = -1


class CarlanetEventListener(abc.ABC):

    def omnet_init_completed(self, run_id, carla_configuration, user_defined) -> (float, SimulatorStatus):
        """
        :param run_id: id corresponding to the one in OMNeT++
        :param carla_configuration:
        :param user_defined:
        :return: current carla world timestamp (see Snapshot class of CarlaAPI), current simulator status
        """
        ...

    def actor_created(self, actor_id: str, actor_type: str, actor_config) -> (float, CarlanetActor):
        """
        Called at the beginning of the simulation, OMNeT says which actors it has and communicate
        with carla to create those actors in the world
        :param actor_id:
        :param actor_type:
        :param actor_config:
        :return: current carla world timestamp (see Snapshot class of CarlaAPI), new actor created from carlaWorld
        """
        ...
        ##return Actor

    def carla_init_completed(self):
        """Called when the initialization of CARLA World is finished"""
        ...

    def carla_simulation_step(self, timestamp) -> SimulatorStatus:
        """
        Method called every time OMNeT call simulation_step of Carla
        :param timestamp
        :return: current simulator status
        """
        ...

    def generic_message(self, timestamp, user_defined_message) -> (SimulatorStatus, dict):
        """
        :param timestamp:
        :param user_defined_message:
        :return: (current simulator status, dict contained custom parameters not None)
        """
        ...

    def simulation_finished(self):
        """
        Callback called upon successful completion of the simulation
        :return:
        """
        ...

    def simulation_error(self, exception):
        """

        :param exception:
        :return:
        """
        ...



from setuptools import find_packages, setup
import sys
from pathlib import Path

setup(
    name="pycarlanet",
    packages=find_packages(include=['pycarlanet', 'pycarlanet.utils']),
    version="0.8.12-alpha",
    long_description=Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type='text/markdown',
    author="Valerio Cislaghi, Christian Quadri",
    license='MIT',
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*,<3.9',
    install_requires=[
        'pyzmq==23.2.1',
        'carla==0.9.13'
    ],
    project_urls={
        'Source': 'https://github.com/jaivra/pycarlanet',
    },
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest==7.2.1'
    ],
    test_suite='tests'
    # setup_requires=['pytest-runner'],
    # tests_require=['pytest==4.4.1'],
    # test_suite='tests',
)
